/*
 * RFBRectangle.h
 *
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

@class NSData;
@class RFBProtocolAdaptor;

#define RFBEncodingRaw			0
#define RFBEncodingCopyRect		1
#define RFBEncodingRRE			2
#define RFBEncodingCoRRE		4
#define RFBEncodingHextile		5

#define RFBHextileRaw			(1 << 0)
#define RFBHextileBackgroundSpecified	(1 << 1)
#define RFBHextileForegroundSpecified	(1 << 2)
#define RFBHextileAnySubrects		(1 << 3)
#define RFBHextileSubrectsColoured	(1 << 4)

@interface RFBRectangle : NSObject <NSCopying, NSCoding>
{
    NXRect _location;
    unsigned _encoding;
}

+ (int *)supportedEncodings;
+ (unsigned)numberOfSupportedEncodings;

+ (RFBRectangle *)rectangleWithAdaptor:(RFBProtocolAdaptor *)adaptor;
+ (RFBRectangle *)rectangleWithLocation:(NXRect)location encoding:(unsigned)encoding adaptor:(RFBProtocolAdaptor *)adaptor;

- (id)initWithLocation:(NXRect)location encoding:(unsigned)encoding adaptor:(RFBProtocolAdaptor *)adaptor;

- (void)drawWithFramebuffer:(VNCFramebuffer *)framebuffer;

- (NXRect)location;
- (unsigned)encoding;

@end

@interface RFBRawRectangle : RFBRectangle <NSCoding>
{
    unsigned _bitsPerPixel;
    NSData *_pixelData;
}

- (NSData *)pixelData;

@end

@interface RFBCopyRectangle : RFBRectangle <NSCoding>
{
    NXPoint _source;
}

- (NXPoint)source;

@end

@interface RFBRRERectangle : RFBRectangle <NSCoding>
{
    unsigned _numberOfRects;
    unsigned _backgroundColor;
    NSMutableData *_rectListCover, *_colorListCover;
}

@end

@interface RFBCoRRERectangle : RFBRRERectangle
{
}

@end

@interface RFBHextileRectangle : RFBRectangle
{
    unsigned _lastBackgroundColor;
    NSMutableArray *_subrectangles;
}

- (void)setLastBackgroundColor:(unsigned)color;
- (unsigned)lastBackgroundColor;

@end

@interface RFBHextileSubrectangle : NSObject <NSCoding>
{
    NXRect _location;
    octet _subEncodingMask;
    unsigned _bitsPerPixel;
    NSData *_pixelData;
    unsigned _backgroundColor;
    unsigned _foregroundColor;
    octet _numberOfRects;
    NSMutableData *_rectListCover, *_colorListCover;
}

- (id)initWithLocation:(NXRect)location superrectangle:(RFBHextileRectangle *)superrect adaptor:(RFBProtocolAdaptor *)adaptor;
- (void)drawWithFramebuffer:(VNCFramebuffer *)framebuffer;

@end
