#import "VNC.h"
#import "vncviewer.h"

NSString *VNCDisplayWillTerminateNotification = @"VNCDisplayWillTerminateNotification";

static void handlePortMessage (int fd, VNCDisplayManager *displayManager);
static void timedUpdateHandler (DPSTimedEntry teNum,
								double now,
								VNCDisplayManager *manager);

@implementation VNCDisplayManager

- (void) addTimedUpdate
{
	if (useFunctions) {
		_updateTimer = DPSAddTimedEntry(0.01, timedUpdateHandler, self, 30);
	}
	else {
		_updateTimer = DPSAddTimedEntry(0.01, timedUpdateHandler, self, 30);
	}
}

- (void) removeTimedUpdate
{
	DPSRemoveTimedEntry (_updateTimer);	
}

- (id)initWithAdaptor:(RFBProtocolAdaptor *)adaptor
{
    NXPoint origin;
	NXRect	rect;
	
    if (self = [super init]) {
        NXRect contentRect = NSMakeRect(0, 0, [adaptor width], [adaptor height]);
        
        _adaptor = [adaptor retain];

/*
        if (![NSBundle loadNibNamed:@"NEXTSTEP_VNCDisplayManager" owner:self])
            [NSException raise:NSInternalInconsistencyException
                        format:@"couldn't load NEXTSTEP_VNCDisplayManager.nib!"];
*/

		if (![NXApp loadNibSection:"VNCDisplayManager.nib" owner:self withNames:NO])
            [NSException raise:NSInternalInconsistencyException
                        format:@"couldn't load NEXTSTEP_VNCDisplayManager.nib!"];

        [window setTitle:[[_adaptor desktopName] cString]];
        [window sizeWindow:contentRect.size.width :contentRect.size.height];
        if ([NXApp mainWindow]) {
			NXRect frame;
			[[NXApp mainWindow] getFrame:&frame];
            origin.x = frame.origin.x + 40;
            origin.y = frame.origin.y - 40;
        }
        else {
            origin.x = 100;
            origin.y = [NXApp mainScreen]->screenBounds.size.height
										- 100 
										- [_adaptor height];
        }
        //[window setFrameOrigin:origin];
        //[window moveTo:origin.x :origin.y];

        inputView = [[[VNCInputView alloc] initWithFrame:contentRect
                                                 adaptor:_adaptor] 
												 autorelease];

	/* balazs.pataki@sztaki.hu: these seems necessary under NS3.3: */
	[inputView setFlipped:YES];
		
        displayView = [[[VNCDisplayView alloc] initWithFrame:contentRect] autorelease];
	dpy = displayView;

        [[statusTextField retain] removeFromSuperview];
        [[window contentView] addSubview:displayView];
        [[window contentView] addSubview:statusTextField];
        [[window contentView] addSubview:inputView];
        [statusTextField release];

        //[window setInitialFirstResponder:inputView];
/*
        [inputView addTrackingRect:[inputView frame]
                             owner:inputView
                          userData:@"" assumeInside:NO];
*/


		DPSAddFD([[_adaptor socket] fileDescriptor], handlePortMessage, self, 15);
		[self addTimedUpdate];
        [self requestFullUpdate:nil];

/*       
        _dataPort = [[NSPort alloc] init];
        [_dataPort setDelegate:self];
        [[NSRunLoop currentRunLoop] addPort:_dataPort forMode:NSDefaultRunLoopMode];

        [self requestFullUpdate:nil];
        
        [NSThread detachNewThreadSelector:@selector(protocolMonitor:)
                                 toTarget:self
                               withObject:_dataPort];

        _updateTimer = [[NSTimer scheduledTimerWithTimeInterval:1.0
                                                         target:self
                                                       selector:@selector(timedUpdateHandler:)
                                                       userInfo:nil
                                                        repeats:YES] retain];
*/
        
        [window makeKeyAndOrderFront:nil];
		
		[inputView getFrame:&rect];
		[window setTrackingRect:&rect
						 inside:NO
						  owner:inputView
							tag:5800
						   left:NO
						  right:NO];
        
		[window makeFirstResponder:inputView];

        [statusTextField setStringValue:"Waiting for update..."];
    }

    return self;
}

- requestFullUpdate:(id)sender
{
 	NXRect rect;
	[displayView getFrame:&rect];
	
	if (useFunctions)
		SendFramebufferUpdateRequest(updateRequestX, updateRequestY,
				      updateRequestW, updateRequestH, False);		
	else
   		[_adaptor sendFramebufferUpdateRequestForRect:rect incremental:NO];
	return self;
}

- (RFBProtocolAdaptor *)adaptor
{
    return _adaptor;
}

- windowWillClose:sender
{
	if ([self performClose])
		return self;
    return nil;
}

- windowWillMiniaturize:sender toMiniwindow:miniwindow
{
    // Now this is just gravy.
    NXImage *image = [[[[displayView framebuffer] framebufferImage] copy] autorelease];
	NXSize  size = {48, 48};
	
	[image setScalable:YES];
	[image setSize:&size];
	
    [window setMiniwindowImage:image];
	return self;
}

- (BOOL)performClose
{
    if (NXRunAlertPanel("Confirm Close",
                        "Really close \"%s\"",
                        "Yes", "No", NULL,
                        [window title]) == NX_ALERTDEFAULT) {
        [self close];
        return YES;
    }
    return NO;
}

- (void)invalidate
{
	
    //[_updateTimer invalidate];
   // [[NSRunLoop currentRunLoop] removePort:_dataPort forMode:NSDefaultRunLoopMode];
    //[_dataPort release];

    [_adaptor release];
    _adaptor = nil;
}

- (void)close
{
    NSNotification *notification = [NSNotification notificationWithName:VNCDisplayWillTerminateNotification object:self];

    [self invalidate];
    [window close];
	[self removeTimedUpdate];
    //[_updateTimer release];

/* No NSNotificationQueue in NS 3.3
    [[NSNotificationQueue defaultQueue] enqueueNotification:notification
                                               postingStyle:NSPostNow];
*/
	[[NSNotificationCenter defaultCenter] postNotification:notification];
}


#if 0
- (void)protocolMonitor:(NSPort *)dataPort
{
    NSAutoreleasePool *threadPool = [[NSAutoreleasePool alloc] init];

    NS_DURING {
        while(_adaptor != nil) {
            NSAutoreleasePool *loopPool = [[NSAutoreleasePool alloc] init];
            RFBMessage *message = [RFBMessage messageWithAdaptor:_adaptor];

            if (message == nil)
                break;

            [dataPort sendData:[NSArchiver archivedDataWithRootObject:message]];
            [loopPool release];
        }
    }
    NS_HANDLER {
        _threadException = [localException retain];
        [dataPort sendData:[NSData data]];
    }
    NS_ENDHANDLER;

    [threadPool release];
    [NSThread exit];
}
#endif /* 0 */

- (void)processMessage:(RFBMessage *)message
{
    switch([message type]) {
        case RFBFramebufferUpdate: {
            NSEnumerator *rectangleEnumerator = [[(id)message rectangles] objectEnumerator];
            RFBRectangle *rectangle;
	    NSString *msg;

            while (rectangle = [rectangleEnumerator nextObject])
                [rectangle drawWithFramebuffer:[displayView framebuffer]];
            
            if ([window isVisible]) {
                [displayView lockFocus];
                [[displayView framebuffer] updateMinimum];
                [displayView unlockFocus];
                [window flushWindow];
            }
	    msg = [NSString stringWithFormat:@"Processed %d rectangles",   
											[[(id)message rectangles] count]];
            [statusTextField setStringValue:[msg cString]];

            break;
        }
            
        case RFBBell:
            NXBeep();		// wow.
            break;
            
        case RFBSetColourMapEntries:
            break;

        case RFBServerCutText:
	    {
    		char *data;
    		NXStream *stream;
    		const char *types[4];
    		int i = 0, length, maxlen;

		types[i++] = NXAsciiPboardType;
            	[NXGeneralPboard declareTypes:types num:i owner:nil];
            	[NXGeneralPboard writeType:NXAsciiPboardType data:[(id)message text]
                                                length:strlen([(id)message text])];
	    }
            break;

        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"unknown message %@", [message description]];
    }
    
}


- (void) handlePortMessage
{
	NSAutoreleasePool *loopPool = [[NSAutoreleasePool alloc] init];
	RFBMessage *message = [RFBMessage messageWithAdaptor:_adaptor];

	if (message == nil)
		return;
	
	[self processMessage:message];
//	[dataPort sendData:[NSArchiver archivedDataWithRootObject:message]];
	[loopPool release];
}

- (void) display
{
	if ([window isVisible]) {
		[displayView lockFocus];
		[[displayView framebuffer] updateMinimum];
		[displayView unlockFocus];
		[window flushWindow];
	}
}

VNCDisplayView *dpy;
int nFramebufferDepth;

void
CopyDataToScreenX(CARD8 *buf, int x, int y, int width, int height)
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	NXRect drawRect;
	NXSetRect (&drawRect, x, y, width, height);
	
	[[dpy framebuffer] compositeRawPixelData:(unsigned int*)buf targetRect:drawRect bitsPerPixel:nFramebufferDepth];
	[dpy display];
	[pool release];
}

void
CopyDataToScreen(CARD8 *buf, int x, int y, int width, int height)
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	NXRect drawRect;
	NXSetRect (&drawRect, x, y, width, height);
	
	[[dpy framebuffer] compositeRawPixelData:(unsigned int*)buf targetRect:drawRect bitsPerPixel:nFramebufferDepth];
	if ([[dpy window] isVisible]) {
		[dpy lockFocus];
		[[dpy framebuffer] updateMinimum];
		[dpy unlockFocus];
		[[dpy window] flushWindow];
	}
	[pool release];
}

static void
handlePortMessage (int fd, VNCDisplayManager *displayManager)
{
	[displayManager removeTimedUpdate];

	if (useFunctions) {
		HandleRFBServerMessage();
		[displayManager display];
		[displayManager addTimedUpdate];
	}
	else {
		[(VNCDisplayManager*)displayManager handlePortMessage];
		[displayManager addTimedUpdate];
	}
}


- (void)timedUpdateHandler
{
	NXRect rect;
	[displayView getFrame:&rect];
    
	if (useFunctions)
		SendIncrementalFramebufferUpdateRequest();
	else
		[_adaptor sendFramebufferUpdateRequestForRect:rect incremental:YES];
}

static void
timedUpdateHandler (
	DPSTimedEntry teNum,
	double 		  now,
	VNCDisplayManager *manager)
{
	if (useFunctions)
		SendIncrementalFramebufferUpdateRequest();
	else
		[manager timedUpdateHandler];
}

- (NSData *)desktopTIFFRepresentation
{
#if 0
    return [[[displayView framebuffer] framebufferImage]
	TIFFRepresentationUsingCompression:NSTIFFCompressionJPEG factor:75.0];
#endif
}

@end
