#import "VNC.h"

#import "vncviewer.h"

#define	KEY_BACKSPACE	0xff08
#define	KEY_TAB		0xff09
#define KEY_ENTER	0xff0d
#define KEY_ESC		0xff1b
#define KEY_INS		0xff63
#define KEY_DEL		0xffff
#define KEY_HOME	0xff50
#define KEY_END		0xff57
#define KEY_PGUP	0xff55
#define KEY_PGDN	0xff56
#define KEY_LEFT	0xff51
#define KEY_UP		0xff52
#define KEY_RIGHT	0xff53
#define KEY_DOWN	0xff54


#define KEY_F1		0xffbe
#define KEY_F2		0xffbf
#define KEY_F3		0xffc0
#define KEY_F4		0xffc1
#define KEY_F5		0xffc2
#define KEY_F6		0xffc3
#define KEY_F7		0xffc4
#define KEY_F8		0xffc5
#define KEY_F9		0xffc6
#define KEY_F10		0xffc7
#define KEY_F11		0xffc8
#define KEY_F12		0xffc9

#if sparc
#define	CODE_BACKSPACE	43
#define	CODE_TAB		53
#define	CODE_ENTER		89
#define	CODE_ESC		29
#define	CODE_INS		44
#define	CODE_DEL		66
#define	CODE_HOME		52
#define	CODE_END		74
#define	CODE_PGUP		96
#define	CODE_PGDN		123
#define	CODE_LEFT		24
#define	CODE_UP			20
#define	CODE_RIGHT		28
#define	CODE_DOWN		27
#define	CODE_F1			0 /* This is the help key ...*/
#define CODE_F2			6
#define CODE_F3			8
#define CODE_F4			10
#define CODE_F5			12
#define CODE_F6			14
#define CODE_F7			16
#define CODE_F8			17	/* That's right!*/
#define CODE_F9			18
#define CODE_F10		7
#define CODE_F11		9
#define CODE_F12		11
#elif m68k
#define	CODE_BACKSPACE	27
#define	CODE_TAB		65
#define	CODE_ENTER		42	/* Numeric: 13 */
#define	CODE_ESC		73
#define	CODE_INS		11	/* Numeric key pad: 0 */
#define	CODE_DEL		12  /* Numeric key pad: . */
#define	CODE_HOME		33	/* Numeric key pad: 7 */
#define	CODE_END		17	/* Numeric key pad: 1 */
#define	CODE_PGUP		35	/* Numeric key pad: 9 */
#define	CODE_PGDN		20	/* Numeric key pad: 3 */
#define	CODE_LEFT		9
#define	CODE_UP			22
#define	CODE_RIGHT		16
#define	CODE_DOWN		15
#elif i386
#define	CODE_BACKSPACE	27
#define	CODE_TAB		65
#define	CODE_ENTER		42	/* Numeric: 13 */
#define	CODE_ESC		73
#define	CODE_INS		11	/* Numeric key pad: 0 */
#define	CODE_DEL		12  /* Numeric key pad: . */
#define	CODE_HOME		33	/* Numeric key pad: 7 */
#define	CODE_END		17	/* Numeric key pad: 1 */
#define	CODE_PGUP		35	/* Numeric key pad: 9 */
#define	CODE_PGDN		20	/* Numeric key pad: 3 */
#define	CODE_LEFT		9
#define	CODE_UP			22
#define	CODE_RIGHT		16
#define	CODE_DOWN		15
#endif

@implementation VNCInputView

- (id)initWithFrame:(NXRect)initFrame adaptor:(RFBProtocolAdaptor *)adaptor
{
    if (self = [super initFrame:&initFrame]) {
        _mouseMask = 0;
        _adaptor = [adaptor retain];
		//[[self window] addToEventMask:NX_MOUSEENTEREDMASK|NX_MOUSEEXITEDMASK];
		//[[self window] addToEventMask:NX_MOUSEMOVEDMASK];
		/*
		[[self window] setEventMask:
				NX_LMOUSEDOWNMASK	|
				NX_LMOUSEUPMASK		|
				NX_RMOUSEDOWNMASK	|
				NX_RMOUSEUPMASK		|
				NX_MOUSEMOVEDMASK	|
				NX_LMOUSEDRAGGEDMASK|
				NX_RMOUSEDRAGGEDMASK|
				NX_MOUSEENTEREDMASK |
				NX_MOUSEEXITEDMASK	|
				NX_KEYDOWNMASK		|
				NX_KEYUPMASK];
		*/
    }

    return self;
}

- (void)dealloc
{
    [_adaptor release];

    [super dealloc];
}

- (BOOL)isFlipped
{
    return YES;
}

- (BOOL)isOpaque
{
    return NO;
}

- (BOOL)acceptsFirstResponder
{
    return YES;
}

- mouseDown:(NXEvent *)theEvent
{
	NXPoint newLoc = theEvent->location;
   	[self convertPoint:&newLoc fromView:nil];

    _mouseMask |= 0x01;
	if (useFunctions)
		SendPointerEvent (newLoc.x, newLoc.y, _mouseMask);
	else
    	[_adaptor sendPointerEventWithButtonMask:_mouseMask location:newLoc];
	return self;
}

- mouseUp:(NXEvent *)theEvent
{
	NXPoint newLoc = theEvent->location;
   	[self convertPoint:&newLoc fromView:nil];

    _mouseMask &= ~(0x01);
	if (useFunctions)
		SendPointerEvent (newLoc.x, newLoc.y, _mouseMask);
	else
		[_adaptor sendPointerEventWithButtonMask:_mouseMask location:newLoc];
	return self;
}

- rightMouseDown:(NXEvent *)theEvent
{
	NXPoint newLoc = theEvent->location;
   	[self convertPoint:&newLoc fromView:nil];

    _mouseMask |= 0x02;
	if (useFunctions)
		SendPointerEvent (newLoc.x, newLoc.y, _mouseMask);
	else
		[_adaptor sendPointerEventWithButtonMask:_mouseMask location:newLoc];
	return self;
}

- rightMouseUp:(NXEvent *)theEvent
{
	NXPoint newLoc = theEvent->location;
   	[self convertPoint:&newLoc fromView:nil];

    _mouseMask &= ~(0x02);
	if (useFunctions)
		SendPointerEvent (newLoc.x, newLoc.y, _mouseMask);
	else
		[_adaptor sendPointerEventWithButtonMask:_mouseMask location:newLoc];
	return self;
}

- mouseMoved:(NXEvent *)theEvent
{
	NXRect myFrame;
	NXPoint newLoc = theEvent->location;
   	[self convertPoint:&newLoc fromView:nil];

	[self getFrame:&myFrame];
	
	if ([self mouse:&newLoc inRect:&myFrame])
		if (useFunctions)
			SendPointerEvent (newLoc.x, newLoc.y, _mouseMask);
		else
			[_adaptor sendPointerEventWithButtonMask:_mouseMask location:newLoc];
	return self;
}

- mouseDragged:(NXEvent *)theEvent
{
	NXPoint newLoc = theEvent->location;
   	[self convertPoint:&newLoc fromView:nil];

	if (useFunctions)
		SendPointerEvent (newLoc.x, newLoc.y, _mouseMask);
	else
		[_adaptor sendPointerEventWithButtonMask:_mouseMask location:newLoc];
	return self;
}

- rightMouseDragged:(NXEvent *)theEvent
{
	NXPoint newLoc = theEvent->location;
   	[self convertPoint:&newLoc fromView:nil];

	if (useFunctions)
		SendPointerEvent (newLoc.x, newLoc.y, _mouseMask);
	else
		[_adaptor sendPointerEventWithButtonMask:_mouseMask location:newLoc];
	return self;
}

- mouseEntered:(NXEvent *)theEvent
{
	[[self window] setEventMask:
				NX_MOUSEMOVEDMASK	|
				NX_LMOUSEDOWNMASK	|
				NX_LMOUSEUPMASK		|
				NX_RMOUSEDOWNMASK	|
				NX_RMOUSEUPMASK		|
				NX_LMOUSEDRAGGEDMASK|
				NX_RMOUSEDRAGGEDMASK|
				NX_MOUSEENTEREDMASK |
				NX_MOUSEEXITEDMASK	|
				NX_KEYDOWNMASK		|
				NX_KEYUPMASK];

//	[[self window] addToEventMask:NX_MOUSEMOVEDMASK];
/*
    [[self window] setAcceptsMouseMovedEvents:YES];
*/
    [[VNCController hollowCursor] push];
	return self;
}

- mouseExited:(NXEvent *)theEvent
{
	[[self window] setEventMask:
				NX_LMOUSEDOWNMASK	|
				NX_LMOUSEUPMASK		|
				NX_RMOUSEDOWNMASK	|
				NX_RMOUSEUPMASK		|
				NX_LMOUSEDRAGGEDMASK|
				NX_RMOUSEDRAGGEDMASK|
				NX_MOUSEENTEREDMASK |
				NX_MOUSEEXITEDMASK	|
				NX_KEYDOWNMASK		|
				NX_KEYUPMASK];
//	[[self window] removeFromEventMask:NX_MOUSEMOVEDMASK];
/*
    [[self window] setAcceptsMouseMovedEvents:NO];
*/
    [NXCursor pop];
	return self;
}

- (int) encodeKeyEvent:(NXEvent *)theEvent
{
	int key = theEvent->data.key.charCode;

#if sparc	
	switch (theEvent->data.key.charSet) {
		case 0:
			switch (theEvent->data.key.charCode) {
				case 127:	key = KEY_BACKSPACE;	break;
				case   9:	key = KEY_TAB;			break;
				case  13:	key = KEY_ENTER;		break;
				case  27:	key = KEY_ESC;			break;
			}
			break;
		
		case 1:
			switch (theEvent->data.key.charCode) {
				case 172:	key = KEY_LEFT;			break;
				case 173:	key = KEY_UP;			break;
				case 174:	key = KEY_RIGHT;		break;
				case 175:	key = KEY_DOWN;			break;
			}
			break;
		
		case 254:
			switch (theEvent->data.key.charCode) {
				case 32:	key = KEY_F1;			break;
				case 33:	key = KEY_F2;			break;
				case 34:	key = KEY_F3;			break;
				case 35:	key = KEY_F4;			break;
				case 36:	key = KEY_F5;			break;
				case 37:	key = KEY_F6;			break;
				case 38:	key = KEY_F7;			break;
				case 39:	key = KEY_F8;			break;
				case 40:	key = KEY_F9;			break;
				case 41:	key = KEY_F10;			break;
				case 42:	key = KEY_F11;			break;
				case 43:	key = KEY_F12;			break;

				case 44:	key = KEY_INS;			break;
				case 45:	key = KEY_DEL;			break;
				case 46:	key = KEY_HOME;			break;
				case 47:	key = KEY_END;			break;
				case 48:	key = KEY_PGUP;			break;
				case 49:	key = KEY_PGDN;			break;
			}
		}
#elif hppa
	switch (theEvent->data.key.charSet) {
		case 0:
			switch (theEvent->data.key.charCode) {
				case 127:	key = KEY_BACKSPACE;	break;
				case   9:	key = KEY_TAB;			break;
				case  13:	key = KEY_ENTER;		break;
				case  27:	key = KEY_ESC;			break;
			}
			break;
		
		case 1:
			switch (theEvent->data.key.charCode) {
				case 172:	key = KEY_LEFT;			break;
				case 173:	key = KEY_UP;			break;
				case 174:	key = KEY_RIGHT;		break;
				case 175:	key = KEY_DOWN;			break;
			}
			break;
		
		case 254:
			switch (theEvent->data.key.charCode) {
				case 32:	key = KEY_F1;			break;
				case 33:	key = KEY_F2;			break;
				case 34:	key = KEY_F3;			break;
				case 35:	key = KEY_F4;			break;
				case 36:	key = KEY_F5;			break;
				case 37:	key = KEY_F6;			break;
				case 38:	key = KEY_F7;			break;
				case 39:	key = KEY_F8;			break;
				case 40:	key = KEY_F9;			break;
				case 41:	key = KEY_F10;			break;
				case 42:	key = KEY_F11;			break;
				case 43:	key = KEY_F12;			break;

				case 44:	key = KEY_INS;			break;
				case 45:	key = KEY_DEL;			break;
				case 46:	key = KEY_HOME;			break;
				case 47:	key = KEY_END;			break;
				case 48:	key = KEY_PGUP;			break;
				case 49:	key = KEY_PGDN;			break;
			}
		}
#elif m68k

	switch (theEvent->data.key.charSet) {
		case 0:
			switch (theEvent->data.key.charCode) {
				case 127:	key = KEY_BACKSPACE;	break;
				case   9:	key = KEY_TAB;			break;
				case  13:	key = KEY_ENTER;		break;
				case  27:	key = KEY_ESC;			break;
				
				/*	Mapping to the numeric key pad
				*/
				case  48:	key = KEY_INS;			break;
				case  46:	key = KEY_DEL;			break;
				case  55:	key = KEY_HOME;			break;
				case  49:	key = KEY_END;			break;
				case  57:	key = KEY_PGUP;			break;
				case  51:	key = KEY_PGDN;			break;
				case  3:	key = KEY_ENTER;		break;
			}
			break;
		
		case 1:
			switch (theEvent->data.key.charCode) {
				case 172:	key = KEY_LEFT;			break;
				case 173:	key = KEY_UP;			break;
				case 174:	key = KEY_RIGHT;		break;
				case 175:	key = KEY_DOWN;			break;
			}
			break;
	}
#elif i386

	switch (theEvent->data.key.charSet) {
		case 0:
			switch (theEvent->data.key.charCode) {
				case 127:	key = KEY_BACKSPACE;	break;
				case   9:	key = KEY_TAB;			break;
				case  13:	key = KEY_ENTER;		break;
				case  27:	key = KEY_ESC;			break;
				
			}
			break;
		
		case 1:
			switch (theEvent->data.key.charCode) {
				case 172:	key = KEY_LEFT;			break;
				case 173:	key = KEY_UP;			break;
				case 174:	key = KEY_RIGHT;		break;
				case 175:	key = KEY_DOWN;			break;
			}
			break;
	}

#endif /* sparc */

	return key;
}

- keyDown:(NXEvent *)theEvent
{
	int flags = theEvent->flags;
	int key = [self encodeKeyEvent:theEvent];
	
	if (useFunctions) {
		if (flags & NX_SHIFTMASK)	// no left/right distinction
			SendKeyEvent (0xffe1, True);
		if (flags & NX_CONTROLMASK)
			SendKeyEvent (0xffe3, True);
		if (flags & NX_COMMANDMASK) // command is mapped to META
			SendKeyEvent (0xffe7, True);
		if (flags & NX_ALTERNATEMASK)
			SendKeyEvent (0xffe9, True);
		
		SendKeyEvent (key, True);
	}
	else {
		if (flags & NX_SHIFTMASK)	// no left/right distinction
			[_adaptor sendKeyEventWithPressedFlag:YES key:0xffe1];
		if (flags & NX_CONTROLMASK)
			[_adaptor sendKeyEventWithPressedFlag:YES key:0xffe3];
		if (flags & NX_COMMANDMASK) // command is mapped to META
			[_adaptor sendKeyEventWithPressedFlag:YES key:0xffe7];
		if (flags & NX_ALTERNATEMASK)
			[_adaptor sendKeyEventWithPressedFlag:YES key:0xffe9];
		
		[_adaptor sendKeyEventWithPressedFlag:YES key:key];
	}

#if 0
#endif /* 0 */
	
#if 0
    NSData *data = [[theEvent charactersIgnoringModifiers] dataUsingEncoding:NSASCIIStringEncoding];
    const octet *charPtr = [data bytes];
    unsigned i = 0;

    if ([theEvent modifierFlags] & NSShiftKeyMask)		// no left/right distinction
        [_adaptor sendKeyEventWithPressedFlag:YES key:0xffe1];
    if ([theEvent modifierFlags] & NSControlKeyMask)
        [_adaptor sendKeyEventWithPressedFlag:YES key:0xffe3];
    if ([theEvent modifierFlags] & NSCommandKeyMask)	// command is mapped to META
        [_adaptor sendKeyEventWithPressedFlag:YES key:0xffe7];
    if ([theEvent modifierFlags] & NSAlternateKeyMask)
        [_adaptor sendKeyEventWithPressedFlag:YES key:0xffe9];
       
    for (i = 0; i < [data length]; ++i)
        [_adaptor sendKeyEventWithPressedFlag:YES key:(unsigned)charPtr[i]];
#endif /* 0 */
	return self;
}

- keyUp:(NXEvent *)theEvent
{
	int flags = theEvent->flags;
	int key = [self encodeKeyEvent:theEvent];

	if (useFunctions) {
		if (flags & NX_SHIFTMASK)	// no left/right distinction
			SendKeyEvent (0xffe1, False);
		if (flags & NX_CONTROLMASK)
			SendKeyEvent (0xffe3, False);
		if (flags & NX_COMMANDMASK) // command is mapped to META
			SendKeyEvent (0xffe7, False);
		if (flags & NX_ALTERNATEMASK)
			SendKeyEvent (0xffe9, False);
		
		SendKeyEvent (key, False);
	}
	else {
		if (flags & NX_SHIFTMASK)	// no left/right distinction
			[_adaptor sendKeyEventWithPressedFlag:NO key:0xffe1];
		if (flags & NX_CONTROLMASK)
			[_adaptor sendKeyEventWithPressedFlag:NO key:0xffe3];
		if (flags & NX_COMMANDMASK) // command is mapped to META
			[_adaptor sendKeyEventWithPressedFlag:NO key:0xffe7];
		if (flags & NX_ALTERNATEMASK)
			[_adaptor sendKeyEventWithPressedFlag:NO key:0xffe9];
		
		[_adaptor sendKeyEventWithPressedFlag:NO key:key];
	}

	
#if 0
	if (flags & NX_SHIFTMASK)	// no left/right distinction
		SendKeyEvent (0xffe1, False);
	if (flags & NX_CONTROLMASK)
		SendKeyEvent (0xffe3, False);
	if (flags & NX_COMMANDMASK) // command is mapped to META
		SendKeyEvent (0xffe7, False);
	if (flags & NX_ALTERNATEMASK)
		SendKeyEvent (0xffe9, False);
		
	SendKeyEvent (theEvent->data.key.charCode, False);
#endif /* 0 */

#if 0
    NSData *data = [[theEvent charactersIgnoringModifiers] dataUsingEncoding:NSASCIIStringEncoding];
    const octet *charPtr = [data bytes];
    unsigned i = 0;

    if ([theEvent modifierFlags] & NSShiftKeyMask)		// no left/right distinction
        [_adaptor sendKeyEventWithPressedFlag:NO key:0xffe1];
    if ([theEvent modifierFlags] & NSControlKeyMask)
        [_adaptor sendKeyEventWithPressedFlag:NO key:0xffe3];
    if ([theEvent modifierFlags] & NSCommandKeyMask)	// command is mapped to META
        [_adaptor sendKeyEventWithPressedFlag:NO key:0xffe7];
    if ([theEvent modifierFlags] & NSAlternateKeyMask)
        [_adaptor sendKeyEventWithPressedFlag:NO key:0xffe9];

    for (i = 0; i < [data length]; ++i)
        [_adaptor sendKeyEventWithPressedFlag:NO key:(unsigned)charPtr[i]];
#endif /* 0 */
	return self;
}


@end
